/**
 *   Programeksempel nr 26 - Strre programeksempel.
 *
 *   Eksemplet viser bruken av svrt mye av det TOTALE pensumet i emnet.
 *   Bl.a vises:  pekere, objekter, arv, virtueller, filhndtering,
 *                string, lambda-funksjoner, og fra STL:
 *                <vector>, <queue>, <list> og <map>, samt <algorithm>.
 *   Sammen med EKS_28.CPP og EKS_29.CPP vises NIVET p en eksamensoppgave
 *   (men dette eksemplet er i OMFANG strre).
 *
 *   Programmet:
 *     - hndterer objekter av typene 'Alpinist' og 'Snowboarder'
 *     - inneholder en heisk (queue), selve skiheisen (queue),
 *       grnn lype (map) og rd lype (list)
 *     - lar objekter flytte seg mellom disse fire ulike containerne
 *     - skriver ut innholdet i de fire containerne
 *     - lar brukeren:   - opprette nye objekter
 *                       - skrive detaljerte data om ETT objekt
 *                       - slette objekter
 *                       - (endre dataene i ETT objekt)
 *     - leser fra og skriver ALLE 4x containerne til fil
 *
 *   @file     EKS_26.CPP
 *   @author   Frode Haug, NTNU
 */

#include <iostream>                //  cout, cin
#include <fstream>                 //  ifstream, ofstream
#include <iomanip>                 //  setw
#include <string>                  //  string
#include <vector>                  //  vector    |
#include <queue>                   //  queue     | (stack det eneste av lrte
#include <list>                    //  list      | containere som IKKE brukes)
#include <map>                     //  map       |
#include <algorithm>               //  algorithm
#include "LesData2.h"              //  Verktykasse for lesing av diverse data
using namespace std;


/**
 *  Korttype (med dagskort, helge-/weekendkort og rskort).
 */
enum KortType  {  Dag,  Helg,  Aar };


/**
 *  Operasjon (med  SE en kjrers DETALJERTE data eller  ENDRE dataene).
 */
enum Operasjon {  Detaljer, Endre  };


//  NB:  I hele eksemplet er det brukt 'G' og 'R' for grnn og rd lype.
//       Dette kunne selvsagt, og like naturlig/bra, vrt erstattet av 'enum'.


/**
 *  Kjrer (med unikt nummer/ID, navn, korttype og kjrte grnne/rde lyper).
 */
class Kjorer {
  private:
      int       nr;
      string    navn;
      KortType  kort;
      vector <char> loyperKjort;

  public:
      Kjorer() { nr = 0; kort = Dag; }    //  Denne constructoren brukes ikke.
      Kjorer(const int nr);
      Kjorer(ifstream & inn);
      virtual ~Kjorer() {  }    //  Tom destructor. Ikke noe  si 'delete' om.

                         //  'hentID' M LAGES I ALLE PROGRAM DER OBJEKTER
                         //   LIGGER I CONTAINERNE <list> (og/eller <map>) !!!
                         //   DENNE RETURNERER KUN KEYEN/DET OBJEKTET SKAL
                         //   SORTERES ETTER:  EN 'int' ELLER EN 'string'.
      int   hentID()    {    return nr;    }
      void  registrerLoype(const char loype);
      virtual void endreData();
      virtual void lesData();
      virtual void skrivData() const;
      virtual void skrivDetaljer() const;
      virtual void skrivTilFil(ofstream & ut) const;
};


/**
 *  Alpinist (med om har twintip ski eller ei, og tightsdress eller ei).
 */
class Alpinist : public Kjorer {
  private:
    bool twinTipSki,
         tightsDress;
  public:
    Alpinist() { twinTipSki = tightsDress = false; }      //  Brukes ikke her.
    Alpinist(const int nr);
    Alpinist(ifstream & inn);
    virtual ~Alpinist() {  }    //  Tom destructor. Ikke noe  si 'delete' om.
    virtual void endreData();
    virtual void lesData();
    virtual void skrivData() const;
    virtual void skrivDetaljer() const;
    virtual void skrivTilFil(ofstream & ut) const;
};


/**
 *  Snowboarder (med om har saggebukse eller ei, og mrk hjelm eller ei).
 */
class Snowboarder : public Kjorer {
  private:
    bool saggeBukse,
         morkHjelm;
  public:
    Snowboarder() {  saggeBukse = morkHjelm = false; }    //  Brukes ikke her.
    Snowboarder(const int nr);
    Snowboarder(ifstream & inn);
    virtual ~Snowboarder() {  }  // Tom destructor. Ikke noe  si 'delete' om.
    virtual void endreData();
    virtual void lesData();
    virtual void skrivData() const;
    virtual void skrivDetaljer() const;
    virtual void skrivTilFil(ofstream & ut) const;
};


void detaljerEllerEndre(const Operasjon operasjon);
Kjorer* finnKjorer(const int nr, char & loype);
void lesFraFil();
void nyKjorer();
void skrivGronn();
void skrivKo(const string & s, queue <Kjorer*> ko);
void skrivMeny();
void skrivQueueTilFil(queue <Kjorer*> ko, ofstream & ut);
void skrivRod();
void skrivTilFil();
void slettAlleData();
void slette();
void tilHeisen();
void tilHeisKo();
void tilLoype(const char loype);


queue <Kjorer*> gHeisKo;       ///<  De som str i heis-k.
queue <Kjorer*> gHeis;         ///<  De som er (p vei opp) i heisen.
list  <Kjorer*> gRod;          ///<  De som kjrer rd lype.
map   <int, Kjorer*> gGronn;   ///<  De som kjrer grnn lype.
int   gSisteNummer;            ///<  Nummer for siste registrerte kjrer.


/**
 *  Hovedprogrammet:
 */
int main()  {
    char kommando;

    lesFraFil();

    skrivMeny();
    kommando = lesChar("Kommando");

    while (kommando != 'Q') {
        switch (kommando) {
          case 'K':    tilHeisKo();                      break;
          case 'H':    tilHeisen();                      break;
          case 'G': case 'R':  tilLoype(kommando);       break;
          case 'N':    nyKjorer();                       break;
          case 'D':    detaljerEllerEndre(Detaljer);     break;
          case 'E':    detaljerEllerEndre(Endre);        break;
          case 'S':    slette();                         break;
          case '1':    skrivKo("heisk", gHeisKo);       break;
          case '2':    skrivKo("heisen", gHeis);         break;
          case '3':    skrivGronn();                     break;
          case '4':    skrivRod();                       break;
          default:     skrivMeny();                      break;
        }
        kommando = lesChar("Kommando");
    }

    skrivTilFil();

    slettAlleData();

    cout << "\n\n";
    return 0;
}


// ---------------------------------------------------------------------------
//                       DEFINISJON AV KLASSE-FUNKSJONER:
// ---------------------------------------------------------------------------

/**
 *  Setter eget nummer/ID (ut fra parameters verdi).
 *
 *  @param   nr  - kjrerens eget nummer/ID
 */
Kjorer::Kjorer(const int nr) {
    kort = Dag;
    this->nr = nr;                  //  M bruke 'this' om egen verdi, nr
}                                   //    parameteren heter EKSAKT det samme.


/**
 *  Leser ALLE egne datamedlemmer inn fra fil.
 *
 *  @param   inn  - Filobjektet egne data leses inn fra
 */
Kjorer::Kjorer(ifstream & inn) {
    char tegn;
    inn >> nr >> tegn;                   //  Leser nummer/ID og 'D/H/A'.
    inn.ignore();                        //  Forkaster ' ' fr navnet/tekst.
    switch (tegn)  {                     //  Konverterer bokstav til enum:
      case 'D':  kort = Dag;   break;
      case 'H':  kort = Helg;  break;
      case 'A':  kort = Aar;   break;
    }
    getline(inn, navn);                  //  Leser eget navn.
}


/**
 *  Tilbyr brukeren  endre p aktuelle datamedlemmer.
 */
void Kjorer::endreData() {
    cout << "\n\tEndre datamedlemmer:  IKKE IMPLEMENTERT ENN\n\n";

    //  LATER
}


/**
 *  Leser inn egne datamedlemmer fra brukeren/tastaturet.
 */
void Kjorer::lesData() {
    char tegn;

    cout << "\tNavn:     ";                //  Leser eget navn:
    getline(cin, navn);

    do {                                   //  Leser ALLTID 'D/H/A':
        tegn = lesChar("\tKorttype (D(ag), H(elg), A(ar)");
    } while (tegn != 'D'  &&  tegn != 'H'  &&  tegn != 'A');

    switch (tegn) {                        //  Konverterer bokstav til enum:
      case 'D':  kort = Dag;    break;
      case 'H':  kort = Helg;   break;
      case 'A':  kort = Aar;    break;
    }
}


/**
 *  Legger til 'G' eller 'A' bakerst i vectoren 'loyperkjort'.
 *
 *  @param   loype  - Lypen kjrt ('G' eller 'R')
 */
void Kjorer::registrerLoype(const char loype) {
    loyperKjort.push_back(loype);        //  Legger ny inn bakerst i vectoren.
}


/**
 *  Skriver ut ALLE egne data p skjermen.
 */
void Kjorer::skrivData() const  {                 //  Skriver nr/ID og navn:
    cout << setw(10) << nr << ":  " << navn << ",  ";
    switch (kort) {                               //  Skriver enum som tekst:
      case Dag:   cout << "dags";    break;
      case Helg:  cout << "helge";   break;
      case Aar:   cout << "rs";     break;
    }                                             //  Antall lypeturer:
    cout << "kort,  antall turer: " << loyperKjort.size() << '\n';
}


/**
 *  Teller opp og skriver antall turer i grnn/rd lype, samt turrekkeflgen.
 */
void Kjorer::skrivDetaljer() const {
    int antG = 0, antR = 0;            //  Nullstiller antall turer i lypene.

    cout << "\tLyper kjrt i dag:\n\t";
    for (const auto val : loyperKjort) {      //  Gr gjennom alle turene:
        if (val == 'G') antG++;  else antR++; //  Teller opp antall i G/R.
        cout << ' ' << val;                   //  Skriver 'G' eller 'R'.
    }
    cout << "\n\tAlts: " << antG << " gang(er) i GRNN, og "
                          << antR << " gang(er) i RD lype.\n\n";
}


/**
 *  Skriver egne data til fil (untatt 'loyperKjort'!).
 *
 *  @param   ut  - Filobjektet egne data skrives ut til
 */
void Kjorer::skrivTilFil(ofstream & ut) const {
    ut << ' ' << nr << ' ';                //  Skriver ' ', eget nr/ID og ' '.
    switch (kort) {                        //  Konverterer og skriver
      case Dag:   ut << 'D';    break;     //    enum som bokstav:
      case Helg:  ut << 'H';    break;
      case Aar:   ut << 'A';    break;
    }
    ut << ' ' << navn << '\n';             //  Skriver ' ' og navnet.
}


// ---------------------------------------------------------------------------

/**
 *  Tom constructor, som bare sender parameter til baseklassen.
 *
 *  @param   nr  - Kjrerens nummer/ID, som sendes til baseklassen
 */
Alpinist::Alpinist(const int nr) : Kjorer(nr) {
    twinTipSki = tightsDress = false;
}


/**
 *  Leser ALLE egne datamedlemmer inn fra fil (etter at baseklassen har lest).
 *
 *  @param   inn  - Filobjektet egne data leses inn fra
 */
Alpinist::Alpinist(ifstream & inn) : Kjorer(inn)  {
    char tegn;
    inn >> tegn;                      //  Leser 2x '+/-', og gjr om til bool:
    twinTipSki = (tegn == '+');
    inn >> tegn;    inn.ignore();
    tightsDress = (tegn == '+');
}


/**
 *  Tilbyr brukeren  endre p aktuelle datamedlemmer.
 *
 *  @see   Kjorer::endreData()
 */
void Alpinist::endreData() {
    Kjorer::endreData();                         //  Baseklassens data endres.

    //  LATER:   Endre datamedlemmer - IKKE IMPLEMENTERT ENN.
}


/**
 *  Leser inn egne datamedlemmer fra brukeren/tastaturet.
 *
 *  @see   Kjorer::lesData()
 */
void Alpinist::lesData() {
    Kjorer::lesData();                            //  Baseklassens data leses.
    twinTipSki = (lesChar("\tTwintip ski?  (j/N)") == 'J');    //  Egne leses:
    tightsDress = (lesChar("\tTights dress? (j/N)") == 'J');
}


/**
 *  Skriver ut ALLE egne data p skjermen.
 *
 *  @see   Kjorer::skrivData()
 */
void Alpinist::skrivData() const  {
    Kjorer::skrivData();                    //  Baseklassens data skrives.
    cout << "\t\tAlpinist "                 //  Egne booler skrives som tekst:
         << (twinTipSki ? "med " : "uten ") << "Twintip ski og "
         << (tightsDress ? "med " : "uten ") << "Tightsdress\n";
}


/**
 *  Skriver ABSOLUTT ALLE datamedlemmer.
 *
 *  @see   skrivData()
 *  @see   Kjorer::skrivDetaljer()
 */
void Alpinist::skrivDetaljer() const {
    skrivData();                     //  Alle andre datadatamedlemmer skrives.
    Kjorer::skrivDetaljer();         //  Baseklassens detaljer skrives.
}


/**
 *  Skriver egne og baseklassens data til fil.
 *
 *  @param   ut  - Filobjektet egne data skrives ut til
 *  @see     Kjorer::skrivTilFil(...)
 */
void Alpinist::skrivTilFil(ofstream & ut) const {
    ut << 'A';                              //  Subklasse av typen 'A'.
    Kjorer::skrivTilFil(ut);                //  Baseklassens data skrives.
                                            //  Egne booler skrives som '+/-':
   ut << (twinTipSki ? "+" : "-") << ' ' << (tightsDress ? "+" : "-") << '\n';
}


// ---------------------------------------------------------------------------

/**
 *  Tom constructor, som bare sender parameter til baseklassen.
 *
 *  @param   nr  - Kjrerens nummer/ID, som sendes til baseklassen
 */
Snowboarder::Snowboarder(const int nr) : Kjorer(nr) {
    saggeBukse = morkHjelm = false;
}


/**
 *  Leser ALLE egne datamedlemmer inn fra fil (etter at baseklassen har lest).
 *
 *  @param   inn  - Filobjektet egne data leses inn fra
 */
Snowboarder::Snowboarder(ifstream & inn) : Kjorer(inn)  {
    char tegn;
    inn >> tegn;                      //  Leser 2x '+/-', og gjr om til bool:
    saggeBukse = (tegn == '+');
    inn >> tegn;    inn.ignore();
    morkHjelm = (tegn == '+');
}


/**
 *  Tilbyr brukeren  endre p aktuelle datamedlemmer.
 *
 *  @see   Kjorer::endreData()
 */
void Snowboarder::endreData() {
    Kjorer::endreData();                         //  Baseklassens data endres.

    //  LATER:   Endre datamedlemmer - IKKE IMPLEMENTERT ENN.
}


/**
 *  Leser inn egne datamedlemmer fra brukeren/tastaturet.
 *
 *  @see   Kjorer::lesData()
 */
void Snowboarder::lesData() {
    Kjorer::lesData();                            //  Baseklassens data leses.
    saggeBukse = (lesChar("\tSaggebukse?  (J/n)") != 'N');     //  Egne leses:
    morkHjelm  = (lesChar("\tMrk hjelm?  (J/n)") != 'N');
}


/**
 *  Skriver ut ALLE egne data p skjermen.
 *
 *  @see   Kjorer::skrivData()
 */
void Snowboarder::skrivData() const  {
    Kjorer::skrivData();                    //  Baseklassens data skrives.
    cout << "\t\tSnowboarder "              //  Egne booler skrives som tekst:
        << (saggeBukse ? "med " : "uten ") << "Saggebukse og "
        << (morkHjelm  ? "med " : "uten ") << "Mrk hjelm\n";
}


/**
 *  Skriver ABSOLUTT ALLE datamedlemmer.
 *
 *  @see   skrivData()
 *  @see   Kjorer::skrivDetaljer()
 */
void Snowboarder::skrivDetaljer() const {
    skrivData();                     //  Alle andre datadatamedlemmer skrives.
    Kjorer::skrivDetaljer();         //  Baseklassens detaljer skrives.
}


/**
 *  Skriver egne og baseklassens data til fil.
 *
 *  @param   ut  - Filobjektet egne data skrives ut til
 *  @see     Kjorer::skrivTilFil(...)
 */
void Snowboarder::skrivTilFil(ofstream & ut) const {
    ut << 'S';                             //  Subklasse av typen 'S'.
    Kjorer::skrivTilFil(ut);               //  Baseklassens data skrives.
                                           //  Egne booler skrives som '+/-':
    ut << (saggeBukse ? "+" : "-") << ' ' << (morkHjelm ? "+" : "-") << '\n';

}


// ---------------------------------------------------------------------------
//                       DEFINISJON AV ANDRE FUNKSJONER:
// ---------------------------------------------------------------------------

/**
 *  Skriver ALLE detaljer ELLER endrer data om EN kjrer.
 *
 *  @param   operasjon  - 'Detaljer' eller 'Endre' - etter hva som skal gjres
 *  @see     finnKjorer(...)
 *  @see     virtual Kjorer::skrivDetaljer()
 *  @see     virtual Kjorer::endreData()
 */
void detaljerEllerEndre(const Operasjon operasjon) {
    Kjorer* kjorer = nullptr;    //  Peker til en Kjorer INNI EN CONTAINER !!!
    char loype;                  //  M vre med som parameter - brukes ikke.
    int nr;                      //  nsket kjrernummer  se/endre.

    if (!gGronn.empty() || !gRod.empty()) {    //  Kjrer(e) i MINST en lype:
        nr = lesInt("\tNr", 1, gSisteNummer);  //  Leser nsket kjrernummer.

        kjorer = finnKjorer(nr, loype);        //  Prver  finne kjreren i
                                               //  en av lypene.
        if (kjorer) {                          //  Kjreren funnet:
            if (operasjon == Detaljer)
                kjorer->skrivDetaljer();       //  Vil se ALLE datadetaljer.
            else
                kjorer->endreData();           //  Vil endre datamedlemmer.
        } else
          cout << "\n\tUkjent nummer i begge lypene!\n\n";
    } else
        cout << "\n\tBegge lypene er tomme!\n\n";
}


/**
 *  Leter etter (og evt. returnerer peker til) en gitt kjrer i en av lypene.
 *
 *  NB: LETER IKKE I DE 2x KENE, DA FOR TUKLETE (OG HELLER IKKE MENINGEN/
 *      POENGET MED EN QUEUE). EI HELLER ER DA ITERATORER TILGJENGELIGE.
 *
 *  @param   nr     -  Kjrernummer det letes etter
 *  @param   loype  -  Evt. funnet i 'G'rnn eller 'R'd lype (REF.OVERFRT)
 *  @return  Peker til funnet 'Kjorer' INNI EN AV CONTAINERNE,  evt. 'nullptr'
 *  @see     Kjorer::hentID()
 */
Kjorer* finnKjorer(const int nr, char & loype) {

    auto it1 = gGronn.find(nr);     //  Prver  finne vha. medlemsfunksjon.
    if (it1 != gGronn.end()) {      //  Funn i GRNN lype (<map>):
        loype = 'G';                //  Setter om hvor funnet.
        return (it1->second);       //  Returnerer peker til 'Kjorer'.
    }                               //  NB:  En 'Kjorer' er tilpekt av
                                    //       'it1->second' !!!

                                    //  Ingen egen "find"-funksjon i <list>,
                                    //    derfor bruke funksjon i <algorithm>,
                                    //    iteratorer og lambda-funksjon:
                                    //  NB:  Bruken av 'hentID()' !!!
    auto it2 = find_if(gRod.begin(), gRod.end(),
                   [nr] (const auto & val) { return(val->hentID() == nr); } );
    if (it2 != gRod.end()) {        //  Funn i RD lype (<list>):
        loype = 'R';                //  Setter om hvor funnet.
        return *it2;                //  Retunerer peker til 'Kjorer'.
    }                               //  NB:  'it2' ER SELV EN PEKER.
                                    //       Derfor ville 'return it2' ha vrt
                                    //       retur av en peker til en peker.
                                    //       'return *it2' blir derfor korrekt
                                    //       return av en peker til 'Kjorer'.

    loype = '-';                    //  Ikke funnet:
    return nullptr;
}


/**
 *  Leser ALLE kjrere inn fra fil og ALLE legges rett inn i heisken.
 */
void lesFraFil() {
    ifstream  innfil("EKS_26.DTA");                 //  pner aktuell fil.
    char kjorerType;                                //  'A' eller 'S'.

    if (innfil) {                                   //  Filen funnet:
        cout << "\n\nLeser fra filen 'EKS_26.DTA' .....\n\n";
        innfil >> gSisteNummer;   innfil.ignore();  //  Leser siste nr.brukt.
        innfil >> kjorerType;                       //  Prver  lese 1.felt.
        while (!innfil.eof()) {       //  Enn ikke slutt p filen:
            switch (kjorerType) {     //  Lager og legger inn aktuell kjrer:
              case 'A': gHeisKo.push(new Alpinist(innfil));            break;
              case 'S': gHeisKo.push(new Snowboarder(innfil));         break;
              default:  cout << "\n\tUlovlig kjrertype p filen!\n";  break;
            }
            innfil >> kjorerType;                   //  Prver  lese 1.felt.
        }
        innfil.close();                             //  Lukker manuelt filen.
    } else
      cout << "\n\nFant ikke filen 'EKS_26.DTA'!\n\n";
}


/**
 *  Legger/setter inn en ny aktuell kjrer i heisken.
 *
 *  @see   virtual Kjorer::lesData()
 */
void nyKjorer() {
    Kjorer* nyKjorer = nullptr;              //  Peker til nye aktuell kjrer.
    char kjorerType;                         //  'A' eller 'S'.
                                //  Teller opp og skriver nytt unikt kjrernr:
    cout << "\tNy kjrer har nr." << ++gSisteNummer << ":\n";

    do {                                     //  Leser ALLTID 'A' eller 'S':
        kjorerType = lesChar("\tA(lpinist) eller S(nowboarder)");
    } while (kjorerType != 'A'  &&  kjorerType != 'S');

    switch (kjorerType) {                    //  Lager en ny aktuell kjrer:
       case 'A':  nyKjorer = new Alpinist(gSisteNummer);     break;
       case 'S':  nyKjorer = new Snowboarder(gSisteNummer);  break;
    }

    nyKjorer->lesData();                   //  Leser kjrerens datamedlemmer.
    gHeisKo.push(nyKjorer);                //  Legger inn den nye kjreren.
}


/**
 *  Skriver ALLE kjrerne i den GRNNE lypen.
 *
 *  @see   virtual Kjorer::skrivData()
 */
void skrivGronn() {
    cout << "\nFlgende " << gGronn.size() << " kjrere er i GRNN lype:\n";
    for (const auto & val : gGronn)        //  Range-based for-lkke:
        (val.second)->skrivData();         //  Skriver kjrerens data.
                                           //  'val.second' er en PEKER!
//  Alternativt vha. iterator, 'for_each(...)' og lambda-funksjon:
//    for_each(gGronn.begin(), gGronn.end(),
//                    [] (const auto & val) { (val.second)->skrivData();  } );

    cout << "\n\n";
}


/**
 *  Skriver HELE en ks innhold.
 *
 *  @param   s  - Ledetekst i utskrift
 *  @param   ko - queue-container med Kjorer-pekere (lagret i en 'deque')
 *                NB: Fr KOPI av PEKERNE til objektene i ken !!!
 *                    Men de TILPEKTE OBJEKTENE er de SAMME !!!
 *                    Hadde det vrt brukt '&', s hadde original-pekerne
 *                    vrt brukt - og blitt borte (nr det sies 'pop') !!!
 *  @see    virtual Kjorer::skrivData()
 */
void skrivKo(const string & s, queue <Kjorer*> ko) {
    cout << "\nFlgende " << ko.size() << " kjrere er i " << s << ":\n";
    while (!ko.empty()) {               //  Enn elementer igjen:
        (ko.front())->skrivData();      //  Viser/skriver den 1.TILPEKTE.
        ko.pop();                       //  Tar ut/fjerner den 1.PEKEREN.
    }
    cout << "\n\n";
}


/**
 *  Skriver programmets menyvalg/muligheter p skjermen.
 */
void skrivMeny() {
    cout << "\nFlgende kommandoer er tilgjengelig:\n"
         << "\tK - inn i heisK\n"
         << "\tH - fra heisk til Heisen\n"
         << "\tG - fra heisen til Grnn lype\n"
         << "\tR - fra heisen til Rd lype\n\n"
         << "\tN - Ny kjrer\n"
         << "\tD - Detaljer om EN kjrer  (KUN mulig nr er i en lype)\n"
         << "\tE - Endre en kjrer        (KUN mulig nr er i en lype)\n"
         << "\tS - Slett/fjern en kjrer  (KUN mulig nr er i en lype)\n\n"
         << "\t1 - skriv HEISK\n"
         << "\t2 - skriv HEISEN\n"
         << "\t3 - skriv GRNN lype\n"
         << "\t4 - skriv RD lype\n\n"
         << "\tQ - Quit / avslutt\n\n";
}


/**
 *  Skriver HELE EN medsendt k til angitt fil.
 *
 *  @param   ko - queue-container med Kjorer-pekere (lagret i en 'deque').
 *                Se ellers NB-kommentaren ifm 'ko' i 'skrivKo'-funksjonen.
 *  @param   ut - Filobjektet ken skal skrives til
 *  @see     virtual Kjorer::skrivTilFil(...)
 */
void skrivQueueTilFil(queue <Kjorer*> ko, ofstream & ut) {
    while (!ko.empty()) {               //  Enn elementer igjen:
        (ko.front())->skrivTilFil(ut);  //  Skriver elementet til fil.
        ko.pop();                       //  Tar ut/fjerner den 1.PEKEREN.
    }
}


/**
 *  Skriver ALLE kjrerne i den RDE lypen.
 *
 *  @see   virtual Kjorer::skrivData()
 */
void skrivRod() {
    cout << "\nFlgende " << gRod.size() << " kjrere er i RD lype:\n";
    for (const auto & val : gRod)             //  Range-based for-lkke:
        val->skrivData();                     //  Skriver kjrerens data.
                                              //  'val' er en PEKER!
//  Alternativt vha. iterator, 'for_each(...)' og lambda-funksjon:
//    for_each(gRod.begin(), gRod.end(),
//                           [] (const auto & val) { val->skrivData();  } );
    cout << "\n\n";
}


/**
 *  Skriver ALLE 4x containernes innhold/kjrere til EN OG SAMME fil.
 *
 *  @see    skrivQueueTilFil(...)
 *  @see    virtual Kjorer::skrivTilFil(...)
 */
void skrivTilFil() {
    ofstream utfil("EKS_26.DT2");                 //  pner aktuell fil.
                                                  //  NB:  'DT2' !
    cout << "\nSkriver til filen 'EKS_26.DT2' .....\n\n";

    utfil << gSisteNummer << '\n';                //  Skriver siste nr. brukt.

    skrivQueueTilFil(gHeis, utfil);               //  Heisens innhold.

    skrivQueueTilFil(gHeisKo, utfil);             //  Heiskens innhold.

    for_each(gGronn.begin(), gGronn.end(),        //  Grnn lypes innhold:
        [&utfil] (const auto & val) {  (val.second)->skrivTilFil(utfil);  } );

    for_each(gRod.begin(), gRod.end(),            //  Rd lypes innhold:
        [&utfil] (const auto & val) {  val->skrivTilFil(utfil);  } );

    utfil.close();                                //  Lukker filen manuelt.

//  NB: NB:  ALLE ELEMENTENE ER FORTSATT UBERRT I ALLE CONTAINERNE !!!

//  NB:  Burde EGENTLIG IKKE ha skrevet til fil de som bare har hatt dagskort,
//       eller om det er sndag og kjreren har hatt helgekort.
//       Men, fin velse  utvide koden (flere steder) med dette ......
}


/**
 *  Sletter ALLE elementene og TILPEKTE kjrere i ALLE containerne.
 *
 *  @see   virtual Kjorer::skrivTilFil(...)
 */
void slettAlleData() {
                   //  Alle de syv innrykkede kodelinjene (til kolonne nr.25)
                   //  er "overfldige", s lenge 'skrivTilFil' (rett ovenfor)
                   //  har gjort dette allerede.
                   //  Ellers mtte/kunne filskriving ha vrt gjort p denne
                   //  mten, samtidig som ALT i ALLE containerne slettes.
                        ofstream utfil("EKS_26.DT3");
                        cout << "\nSkriver til filen 'EKS_26.DT3' .....\n\n";
                        utfil << gSisteNummer << '\n';

                                              //  HEISKEN:
    while (!gHeisKo.empty()) {                //  Enn elementer igjen:
                        (gHeisKo.front())->skrivTilFil(utfil);
        delete gHeisKo.front();               //  Sletter 1.TILPEKTE 'Kjorer'.
        gHeisKo.pop();                        //  Fjerner den 1.PEKEREN.
    }
    cout << "\tAntall elementer i HEISKEN:     " << gHeisKo.size() << '\n';

                                              //  HEISEN:
    while (!gHeis.empty()) {                  //  Enn elementer igjen:
                        (gHeis.front())->skrivTilFil(utfil);
        delete gHeis.front();                 //  Sletter 1.TILPEKTE 'Kjorer'.
        gHeis.pop();                          //  Fjerner den 1.PEKEREN.
    }
    cout << "\tAntall elementer i HEISEN:       " << gHeis.size() << '\n';


    while (!gGronn.empty()) {                 //  GRNN LYPE:
        Kjorer* it = (gGronn.begin())->second;  //  PEKER til 1.KJORER.
                        it->skrivTilFil(utfil);
        delete it;                            //  Sletter 1.TILPEKTE 'Kjorer'.
        gGronn.erase(gGronn.begin());         //  Sletter PARET key og PEKER.
//  Alternativt for de fire siste linjene ovenfor:
//      auto it = gGronn.begin();
//                        (it->second)->skrivTilFil(utfil);
//      delete (it->second);                  //  Sletter TILPEKTE 'Kjorer'.
//      gGronn.erase(it);                     //  Sletter PARET key og PEKER.
    }
    cout << "\tAntall elementer i GRNN LYPE:  " << gGronn.size() << '\n';

                                              //  RD LYPE:
    while (!gRod.empty()) {                   //  Enn elementer igjen:
                        (gRod.front())->skrivTilFil(utfil);
        delete gRod.front();                  //  Sletter 1.TILPEKTE 'Kjorer'.
        gRod.pop_front();                     //  Fjerner den 1.PEKEREN.
    }
    cout << "\tAntall elementer i RD LYPE:    " << gRod.size() << '\n';
}


/**
 *  Sletter (om mulig) en nsket nummerangitt kjrer (nr er i en av lypene).
 *
 *  @see   finnKjorer(...)
 *  @see   Kjorer::hentID()
 *  @see   virtual Kjorer:skrivData()
 */
void slette() {
//  NB: Denne funksjonen er MYE LIK 'tilHeisKo'. S, de kunne til en viss grad
//      ha vrt laget p samme mte som 'detaljerEllerEndre'. Men, det hadde
//      blitt spass mange 'if' og 'else' for  skille p meldinger og hva
//      som skal gjres, at derfor er det laget to ulike funksjoner.

    Kjorer* kjorer = nullptr; //  Peker til en Kjorer INNI EN CONTAINER !!!
    char loype;               //  Hvilken lype funnet i (evt. 'G' eller 'R').
    int nr;                   //  nsket kjrernummer  slette.

    if (!gGronn.empty() || !gRod.empty()) {       //  Kjrer i minst EN lype:
       nr = lesInt("\tSlette nr", 0, gSisteNummer);   //  nsket nr.  slette.

       if (nr) {                            // 'nr' er IKKE 0 (null):
          kjorer = finnKjorer(nr, loype);   //  Prver  finne kjreren i
                                            //    en av lypene.
          if (kjorer) {                     //  Kjreren funnet:
             if (loype == 'G')              //  Funnet i GRNN lype:
                gGronn.erase(nr);           //  Sletter PARET key og PEKER.
             else if (loype == 'R') {       //  Er i RD lype:
                                            //  Bruker funksjon i <algorithm>:
                auto it = find_if(gRod.begin(), gRod.end(),
                     [nr](const auto & val) { return(val->hentID() == nr); });
                if (it != gRod.end())       //  Fortsatt funn:
                   gRod.erase(it);          //  Sletter SELVE PEKEREN.
             }
             cout << "\tSlettet er alts:\n";
             kjorer->skrivData();           //  Skriver den som slettes.
             delete kjorer;          //  N FRST SLETTES SELVE KJREREN !!!
                                     //  NB:  Hadde det vrt gjort tidligere
                                     //       (sammen med 'skrivData'), f.eks.
                                     //       rett etter 'if (kjorer)', s
                                     //       ville setningen 'val->hentID()'
                                     //       ha blitt meningsls.
          } else
            cout << "\n\tUkjent nummer i begge lypene!\n\n";
       } else
         cout << "\n\tOK - ingen slettes.\n\n";
    } else
      cout << "\n\tBegge lypene er tomme!\n\n";
}


/**
 *  Flytter (om mulig) 1.kjrer i heisken til inn bakerst/nederst i heisen.
 *
 *  @see   virtual Kjorer::skrivData()
 */
void tilHeisen() {
  if (!gHeisKo.empty()) {                   //  Noen i heisken:
      (gHeisKo.front())->skrivData();       //  Skriver den frste sine data.
    cout << "\n\tflyttes fra heisken og inn (nederst/bakerst) i heisen.\n\n";
      gHeis.push(gHeisKo.front());          //  KOPIERER pekeren til 'gHeis'.
      gHeisKo.pop();                        //  Sletter 1.pekeren i 'gHeisKo'.
  }  else
      cout << "\n\tHeisken er tom, s ingen  sette inn i heisen!\n\n";
}


/**
 *  Flytter (om mulig) en kjrer fra en av lypene til bakerst i heisken.
 *
 *  @see   finnKjorer(...)
 *  @see   virtual Kjorer::skrivData()
 *  @see   Kjorer::hentID()
 */
void tilHeisKo()  {
    Kjorer* kjorer = nullptr;    //  Kjreren evt. funnet i en av lypene.
    char loype;                  //  Evt. funnet i 'G'rnn eller 'R'd lype.
    int nr;                      //  nsket kjrernummer fra lype til heisk.

    if (!gGronn.empty()  ||  !gRod.empty()) {     //  Kjrer i minst EN lype:
        cout << "\tFlytte fra en lype til heisk:\n";
        nr = lesInt("\tNr", 1, gSisteNummer);     //  nsket nr.  flytte.

        kjorer = finnKjorer(nr, loype);       //  Prver  finne kjreren i
                                              //    en av lypene.
        if (kjorer) {                         //  Kjreren funnet:
            kjorer->skrivData();              //  Skriver kjrerens data.
            if (loype == 'G') {               //  Funnet i GRNN lype:
               cout << "\n\tflyttes FRA GRNN lype og inn i heisken.\n\n";
               gHeisKo.push(kjorer);          //  Legger inn Kjorer-PEKER.
               gGronn.erase(nr);              //  Sletter PARET key og PEKER.
           } else if (loype == 'R') {         //  Funnet i RD lype:
               cout << "\n\tflyttes FRA RD lype og inn i heisken.\n\n";
               gHeisKo.push(kjorer);          //  Legger inn Kjorer-PEKER.
                                              //  Sletter SELVE PEKEREN:
               gRod.remove_if([nr] (const auto & val)
                                   { return(val->hentID() == nr); });
//  Alternativt vha. funksjon i <algorithm>, da ingen egen "find" i <list>:
/*       auto it = find_if(gRod.begin(), gRod.end(),
                   [nr] (const auto & val) { return(val->hentID() == nr); } );
         if (it != gRod.end())           //  Funn:
            gRod.erase(it);              //  Sletter SELVE PEKEREN.
         else   //  Kjreren er jo funnet - s what???:
            cout << "\n\tSEMANTISK FEIL-1 - dette skal ikke skje!!\n\n";  */
           } else     //  Kjreren er jo funnet, men IKKE 'G/R' - s what???:
             cout << "\n\tSEMANTISK FEIL-2 - dette skal ikke skje!!\n\n";
                      // *****************************************************
                      // **   Jfr/vs. EXCEPTIONS kap.16.1 i lreboka        **
                      // **                      (som ikke er pensum).      **
                      // *****************************************************
        } else
          cout << "\n\tUkjent nummer i begge lypene!\n\n";
    } else
      cout << "\n\tBegge lypene er tomme!\n\n";
}


/**
 *  Flytter (om mulig) fra heisen og over til grnn eller rd lype.
 *
 *  @param   loype  - Bokstaven ('G/R') for hvilken lype  flytte til
 *  @see     virtual Kjorer::skrivData()
 *  @see     Kjorer::registrerLoype
 *  @see     Kjorer::hentID()
 */
void tilLoype(const char loype) {
  if (!gHeis.empty()) {                          //  Kjrer(e) i heisen:
     (gHeis.front())->skrivData();               //  1.kjrers data skrives.
     cout << "\n\tflyttes fra heisen og over i "
          << ((loype == 'G') ? "GRNN" : "RD") << " lype\n\n";

     gHeis.front()->registrerLoype(loype);       //  Oppdaterer kjrerens
                                                 //    "statistikk" med 'G/R'.
     if (loype == 'G')                           //  Flyttes til GRNN:
                                             //  NKKEL OG PEKER LEGGES INN:
                                             //  NB:  BRUKEN AV 'hentID()' !!!
       gGronn[gHeis.front()->hentID()] = gHeis.front();
//  Alternativt litt mer omstendelig/formelt:
//                gGronn.insert(pair <int, Kjorer*>
//                             ((gHeis.front())->hentID(), gHeis.front()));


     else {                                   //  Flyttes til RD:
       gRod.push_back(gHeis.front());   //  Legger inn PEKER bakerst i listen:
                                        //  NB:  SORTERER !!!!!!!!!!!!!! :
       gRod.sort([](Kjorer* k1, Kjorer* k2)
                                { return (k1->hentID() < k2->hentID()); });
     }
     gHeis.pop();                             //  Fjerner PEKER fra heisen.

  } else
    cout << "\n\tHeisen er tom, s ingen  sette inn i noen lype!\n\n";
}
